@@ -33,6 +33,7 @@
 #include "gda-dict.h"
 #include "gda-parameter-list.h"
 #include "gda-value.h"
+#include "gda-enum-types.h"
 
 typedef struct {
 	GModule              *handle;
@@ -183,8 +184,8 @@ gda_client_class_init (GdaClientClass *klass)
 			      G_SIGNAL_RUN_LAST,
 			      G_STRUCT_OFFSET (GdaClientClass, event_notification),
 			      NULL, NULL,
-			      gda_marshal_VOID__POINTER_INT_POINTER,
-			      G_TYPE_NONE, 3, G_TYPE_POINTER, G_TYPE_INT, G_TYPE_POINTER);
+			      gda_marshal_VOID__OBJECT_ENUM_OBJECT,
+			      G_TYPE_NONE, 3, GDA_TYPE_CONNECTION, GDA_TYPE_CLIENT_EVENT, GDA_TYPE_PARAMETER_LIST);
 
 	object_class->finalize = gda_client_finalize;
 }
@@ -285,6 +286,7 @@ find_or_load_provider (GdaClient *client, const gchar *provider)
 {
 	LoadedProvider *prv = NULL;
 	GdaProviderInfo *prv_info;
+	void (*plugin_init) (const gchar *);
 
 	prv_info = gda_config_get_provider_by_name (provider);
 	if (!prv_info) {
@@ -306,6 +308,15 @@ find_or_load_provider (GdaClient *client, const gchar *provider)
 	
 	g_module_make_resident (prv->handle);
 	
+	/* initialize plugin if supported */
+	if (g_module_symbol (prv->handle, "plugin_init", (gpointer) &plugin_init)) {
+		gchar *dirname;
+
+		dirname = g_path_get_dirname (prv_info->location);
+		plugin_init (dirname);
+		g_free (dirname);
+	}
+
 	g_module_symbol (prv->handle, "plugin_get_name",
 			 (gpointer) &prv->plugin_get_name);
 	g_module_symbol (prv->handle, "plugin_get_description",
